/* -*- mode: c; tab-width: 800; c-basic-offset: 4; -*-
 * ====================================================================
 * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

#include <sys/param.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "suexec.h"

#define ENVBUFLEN 256

extern char **environ;
static FILE *log = NULL;

static struct safe_env {
    char        *key;
    int          len;
    int          strip;
} safe_env_lst[] = {
    { "AUTH_TYPE",                      0, 0 },
    { "CONTENT_LENGTH",                 0, 0 },
    { "CONTENT_TYPE",                   0, 0 },
    { "DATE_GMT",                       0, 0 },
    { "DATE_LOCAL",                     0, 0 },
    { "DOCUMENT_NAME",                  0, 0 },
    { "DOCUMENT_PATH_INFO",             0, 0 },
    { "DOCUMENT_ROOT",                  0, 1 },
    { "DOCUMENT_URI",                   0, 0 },
    { "FILEPATH_INFO",                  0, 0 },
    { "GATEWAY_INTERFACE",              0, 0 },
    { "LAST_MODIFIED",                  0, 0 },
    { "PATH_INFO",                      0, 0 },
    { "PATH_TRANSLATED",                0, 0 },
    { "QUERY_STRING",                   0, 0 },
    { "QUERY_STRING_UNESCAPED",         0, 0 },
    { "REMOTE_ADDR",                    0, 0 },
    { "REMOTE_HOST",                    0, 0 },
    { "REMOTE_IDENT",                   0, 0 },
    { "REMOTE_PORT",                    0, 0 },
    { "REMOTE_USER",                    0, 0 },
    { "REDIRECT_QUERY_STRING",          0, 0 },
    { "REDIRECT_STATUS",                0, 0 },
    { "REDIRECT_URL",                   0, 0 },
    { "REQUEST_METHOD",                 0, 0 },
    { "REQUEST_URI",                    0, 0 },
    { "SCRIPT_FILENAME",                0, 1 },
    { "SCRIPT_NAME",                    0, 0 },
    { "SCRIPT_URI",                     0, 0 },
    { "SCRIPT_URL",                     0, 0 },
    { "SERVER_ADMIN",                   0, 0 },
    { "SERVER_NAME",                    0, 0 },
    { "SERVER_PORT",                    0, 0 },
    { "SERVER_PROTOCOL",                0, 0 },
    { "SERVER_SOFTWARE",                0, 0 },
    { "UNIQUE_ID",                      0, 0 },
    { "USER_NAME",                      0, 0 },
    { "TZ",                             0, 0 },
    { NULL,                             0, 0 },
};

extern char *__progname;

static void
html_err(char *cmd, char *msg, ...)
{
    printf("Content-Type: text/html\n"
           "\n"
           "<html>\n"
           "<head><title>Script error</title></head>\n"
           "<body bgcolor=white text=red>\n"
           "<h1>Script Error</h1>\n"
           "An error was encountered while executing the script <tt>%s</tt>.\n"
           "The error was:\n"
           "<pre>\n"
           "%s"
           "</pre>\n"
           "</body>\n"
           "</html>\n", cmd, msg);
    exit(0);
}

static void
err_openlog(void)
{
#ifdef SUEXEC_LOG
    int fd;

    if ((log = fopen(SUEXEC_LOG, "a")) == NULL) {
        perror("err_openlog: fopen");
        exit(1);
    }

    fd = fileno(log);

    if (fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | FD_CLOEXEC) != 0) {
        perror("err_openlog: fcntl");
        exit(1);
    }
}

static void
err_output(int err, const char *fmt, va_list ap)
{
    time_t timevar;
    struct tm *lt;

    if (!log)
        err_openlog();

    time(&timevar);
    lt = localtime(&timevar);

    fprintf(log, "%s [%d-%.2d-%.2d %.2d:%.2d:%.2d]: ",
            __progname,
            lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday,
            lt->tm_hour, lt->tm_min, lt->tm_sec);

    vfprintf(log, fmt, ap);

    if (err) {
        fprintf(log, ": %s", strerror(errno));
    }

    fputc('\n', log);

    fflush(log);
#endif /* SUEXEC_LOG */
    return;
}

static void
log_err(const char *fmt, ...)
{
#ifdef SUEXEC_LOG
    va_list ap;

    va_start(ap, fmt);
    err_output(1, fmt, ap);
    va_end(ap);
#endif /* SUEXEC_LOG */
    return;
}

static void
log_errx(const char *fmt, ...)
{
#ifdef SUEXEC_LOG
    va_list ap;

    va_start(ap, fmt);
    err_output(0, fmt, ap);
    va_end(ap);
#endif /* SUEXEC_LOG */
    return;
}

static void
clean_env(char *root)
{
    struct safe_env *se;
    char pathbuf[512];
    char **cleanenv;
    char **ep;
    int cidx;

    cidx = 0;

    if ((cleanenv = (char **) calloc(ENVBUFLEN, sizeof(char *))) == NULL) {
        log_err("failed to malloc memory for environment");
        exit(120);
    }

    sprintf(pathbuf, "PATH=%s", SAFE_PATH);
    cleanenv[cidx++] = strdup(pathbuf);

    for (ep = environ; *ep && cidx < ENVBUFLEN-1; ep++) {
        if (!strncmp(*ep, "HTTP_", 5)
            || !strncmp(*ep, "HTTPS", 5)
            || !strncmp(*ep, "SSL_", 4)) {
            cleanenv[cidx++] = *ep;
        } else {
            for (se = safe_env_lst; se->key != NULL; ++se) {
                if (se->len == 0)
                    se->len = strlen(se->key);
                if ((*ep)[se->len] == '=' &&
                    strncmp(*ep, se->key, se->len) == 0) {
#ifdef SUEXEC_CHROOT
                    /*
                     * If root is a prefix of the value of this
                     * variable, strip it off so the path is relative
                     * to the root of the chroot tree.
                     */
                    if (se->strip) {
                        char *p, *q, *r;

                        p = *ep + se->len + 1;
                        for (q = p, r = root; *r && *q == *r; ++q, ++r)
                            /* nothing */ ;
                        if (*r == '\0' && (*q == '\0' || *q == '/')) {
                            while (*q)
                                *p++ = *q++;
                            *p = '\0';
                        }
                    }
#endif
                    cleanenv[cidx++] = *ep;
                    break;
                }
            }
        }
    }

    cleanenv[cidx] = NULL;

    environ = cleanenv;
}

int
main(int argc, char *argv[])
{
#ifndef SUEXEC_CHROOT
    int userdir = 0;            /* ~userdir flag             */
#endif
    uid_t uid;                  /* user information          */
    gid_t gid;                  /* target group placeholder  */
    char *target_uname;         /* target user name          */
    char *target_gname;         /* target group name         */
    char *target_homedir;       /* target home directory     */
    char *actual_uname;         /* actual user name          */
    char *actual_gname;         /* actual group name         */
    char *prog;                 /* name of this program      */
    char *cmd;                  /* command to be executed    */
    char *p;                    /* scratch */
    char cwd[MAXPATHLEN];       /* current working directory */
    struct passwd *pw;          /* password entry holder     */
    struct group *gr;           /* group entry holder        */
    struct stat dir_info;       /* directory info holder     */
    struct stat prg_info;       /* program info holder       */
#ifndef SUEXEC_CHROOT
    char dwd[MAXPATHLEN];       /* docroot working directory */
#endif

    /*
     * If there are a proper number of arguments, set all of them to
     * variables. Otherwise, error out.
     */
    prog = argv[0];
    if (argc < 4) {
        log_errx("incorrect number of arguments");
        exit(101);
    }
    target_uname = argv[1];
    target_gname = argv[2];
#ifdef SUEXEC_CHROOT
    /*
     * Dirty hack to circumvene the fact that Apache trims off
     * everything except the actual file name.
     */
    if ((cmd = getenv("SCRIPT_NAME")) == NULL) {
        log_errx("warning: SCRIPT_NAME not set");
        cmd = argv[3];
    }
#else
    cmd = argv[3];
#endif

    /*
     * Log the transaction here to be sure we have an open log before
     * we setuid() (and possibly chroot()).
     */
    log_errx("invoked for user '%s', group '%s', command '%s'"
#ifdef SUEXEC_CHROOT
             " (chroot)"
#endif
             "",
             target_uname, target_gname, cmd);

    /*
     * Check existence/validity of the UID of the user running this
     * program. Error out if invalid.
     */
    uid = getuid();
    if ((pw = getpwuid(uid)) == NULL) {
        log_errx("invalid uid: (%ld)", uid);
        exit(102);
    }

    /*
     * Check to see if the user running this program is the user
     * allowed to do so as defined in suexec.h. If not the allowed
     * user, error out.
     */
    if (strcmp(HTTPD_USER, pw->pw_name)) {
        log_errx("user mismatch (%s instead of %s)", pw->pw_name, HTTPD_USER);
        exit(103);
    }

#ifndef SUEXEC_CHROOT
    /*
     * Check for a leading '/' (absolute path) in the command to be
     * executed, or attempts to back up out of the current directory,
     * to protect against attacks. If any are found, error out.
     * Naughty naughty crackers.
     */
    if (cmd[0] == '/' || strncmp(cmd, "../", 3) == 0
        || strstr(cmd, "/../") != NULL) {
        log_errx("invalid command (%s)", cmd);
        exit(104);
    }
#endif

    /*
     * Check to see if this is a ~userdir request. If so, set the
     * flag, and remove the '~' from the target username, unless we're
     * chrooting, in which case we just bail out.
     */
    if (!strncmp("~", target_uname, 1)) {
#ifdef SUEXEC_CHROOT
        log_errx("cannot use userdir \"%s\" while chrooting", target_uname);
        exit(122);
#else
        target_uname++;
        userdir = 1;
#endif
    }

    /*
     * Error out if the target username is invalid.
     */
    if ((pw = getpwnam(target_uname)) == NULL) {
        log_errx("invalid target user name: (%s)", target_uname);
        exit(105);
    }

    /*
     * Error out if the target group name is invalid.
     */
    if (strspn(target_gname, "1234567890") != strlen(target_gname)) {
        if ((gr = getgrnam(target_gname)) == NULL) {
            log_errx("invalid target group name: (%s)", target_gname);
            exit(106);
        }
        gid = gr->gr_gid;
        actual_gname = strdup(gr->gr_name);
    } else {
        gid = atoi(target_gname);
        actual_gname = strdup(target_gname);
    }


    /*
     * Save these for later since initgroups will hose the struct
     */
    uid = pw->pw_uid;
    actual_uname = strdup(pw->pw_name);
    target_homedir = strdup(pw->pw_dir);

    /*
     * Strip trailing slashes from target_homedir, but don't let it
     * get empty (i.e. don't strip the slash if target_homedir is "/")
     */
    p = strrchr(target_homedir, '/');
    if (p != NULL && p[1] == '\0')
        while (p > target_homedir && *p == '/')
            *p-- = '\0';

    /*
     * Error out if attempt is made to execute as root or as a UID
     * less than UID_MIN. Tsk tsk.
     */
    if (uid == 0 || uid < UID_MIN) {
        log_err("cannot run as forbidden uid (%d/%s)", uid, cmd);
        exit(107);
    }

    /*
     * Error out if attempt is made to execute as root group or as a
     * GID less than GID_MIN. Tsk tsk.
     */
    if (gid == 0 || gid < GID_MIN) {
        log_err("cannot run as forbidden gid (%d/%s)", gid, cmd);
        exit(108);
    }

#ifdef SUEXEC_CHROOT
    /*
     * If we are attempting to run in a chrooted space, change root
     * here. Error out if we're not able to change both root and dir.
     */
    if (chroot(target_homedir) != 0) {
        log_err("unable to chroot to %s", target_homedir);
        exit(123);
    }
    if (chdir(SUEXEC_CHDIR) != 0) {
        log_err("unable to chdir to " SUEXEC_CHDIR " in chroot dir %s",
                target_homedir);
        exit(124);
    }
#endif

    /*
     * Change UID/GID here so that the following tests work over NFS.
     *
     * Initialize the group access list for the target user, and
     * setgid() to the target group. If unsuccessful, error out.
     */
    if (setgid(gid) != 0 || initgroups(actual_uname, gid) != 0) {
        log_err("failed to setgid (%ld: %s)", gid, cmd);
        exit(109);
    }

    /*
     * setuid() to the target user. Error out on fail.
     */
    if (setuid(uid) != 0) {
        log_err("failed to setuid (%ld: %s)", uid, cmd);
        exit(110);
    }

#ifndef SUEXEC_CHROOT
    /*
     * Get the current working directory, as well as the proper
     * document root (dependant upon whether or not it is a ~userdir
     * request). Error out if we cannot get either one, or if the
     * current working directory is not in the docroot. Use chdir()s
     * and getcwd()s to avoid problems with symlinked directories.
     * Kinda yucky, but it beats using realpath().
     */

    if (getcwd(cwd, MAXPATHLEN) == NULL) {
        log_err("cannot get current working directory");
        exit(111);
    }

    if (userdir) {
        if (chdir(target_homedir) != 0 ||
            chdir(USERDIR_SUFFIX) != 0 ||
            getcwd(dwd, MAXPATHLEN) == NULL ||
            chdir(cwd) != 0) {
            log_err("cannot get docroot information (%s)", target_homedir);
            exit(112);
        }
    } else {
        if (chdir(DOC_ROOT) != 0
            || getcwd(dwd, MAXPATHLEN) == NULL
            || chdir(cwd) != 0) {
            log_err("cannot get docroot information (%s)", DOC_ROOT);
            exit(113);
        }
    }

    if (strncmp(cwd, dwd, strlen(dwd)) != 0) {
        log_err("command not in docroot (%s)", cmd);
        exit(114);
    }
#else
    /*
     * We are running chrooted; there is no reason to do all the
     * magic, as creating references out of the chrooted space is not
     * possible anyway.
     */
    if (getcwd(cwd, MAXPATHLEN) == NULL) {
        log_err("cannot get current working directory");
        exit(111);
    }
#endif

    /*
     * Stat the cwd and verify it is a directory, or error out.
     */
    if (lstat(cwd, &dir_info) != 0 || !S_ISDIR(dir_info.st_mode)) {
        log_err("cannot stat directory %s", cwd);
        html_err(cmd, "Permission denied");
        exit(115);
    }

    /*
     * Error out if cwd is writable by others.
     */
    if (dir_info.st_mode & S_IWOTH) {
        log_errx("directory '%s' is writable by others", cwd);
        html_err(cmd, "Permission denied");
        exit(116);
    }

    /*
     * Error out if cwd is writable by untrusted group.
     */
    if (dir_info.st_mode & S_IWGRP) {
#ifdef SUEXEC_TRUSTED_GID
        if (dir_info.st_gid != SUEXEC_TRUSTED_GID) {
#endif
            log_errx("directory '%s' is writable by group %d",
                     cwd, dir_info.st_gid);
            html_err(cmd, "Permission denied");
            exit(116);
#ifdef SUEXEC_TRUSTED_GID
        }
#endif
    }

    /*
     * Error out if we cannot stat the program.
     */
#ifndef SUEXEC_CHROOT
    if (lstat(cmd, &prg_info) != 0 || S_ISLNK(prg_info.st_mode)) {
        log_err("cannot stat script '%s'", cmd);
        html_err(cmd, "Permission denied");
        exit(117);
    }
#else
    /*
     * Symlinks are not dangerous when we are running in a chrooted
     * environment, so use stat here instead.
     */
    if (stat(cmd, &prg_info) != 0) {
        log_err("cannot stat script '%s'", cmd);
        html_err(cmd, "Permission denied");
        exit(117);
    }
#endif

    /*
     * Error out if the program is writable by others.
     */
    if (prg_info.st_mode & S_IWOTH) {
        log_errx("script '%s' is writable by others", cmd);
        html_err(cmd, "Permission denied");
        exit(118);
    }

    /*
     * Error out if the program is writable by untrusted group.
     */
    if (prg_info.st_mode & S_IWGRP) {
#ifdef SUEXEC_TRUSTED_GID
        if (prg_info.st_gid != SUEXEC_TRUSTED_GID) {
#endif
            log_errx("script '%s' is writable by group %d",
                     cwd, prg_info.st_gid);
            html_err(cmd, "Permission denied");
            exit(118);
#ifdef SUEXEC_TRUSTED_GID
        }
#endif
    }

    /*
     * Error out if the file is setuid or setgid.
     */
    if ((prg_info.st_mode & S_ISUID) || (prg_info.st_mode & S_ISGID)) {
        log_errx("script '%s' is setuid and/or setgid", cmd);
        html_err(cmd, "Permission denied");
        exit(119);
    }

    /*
     * Error out if the target user/group is different from the
     * user/group of the current working directory.
     */
    if (dir_info.st_uid && dir_info.st_uid != uid) {
        log_errx("target uid (%ld) does not match directory (%ld)",
                 uid, dir_info.st_uid);
        html_err(cmd, "Permission denied");
        exit(120);
    }

#if 0
    /*
     * Error out if the target user/group is different from the
     * user/group of the program.
     */
    if (prg_info.st_uid && prg_info.st_uid != uid) {
        log_errx("target uid (%ld) does not match script (%ld)",
                 uid, prg_info.st_uid);
        html_err(cmd, "Permission denied");
        exit(120);
    }
#endif

    clean_env(target_homedir);

    /*
     * Execute the command, replacing our image with its own.
     */
    execv(cmd, &argv[3]);

    /*
     * Uh oh. Still here. Where's the kaboom? There was supposed to be
     * an EARTH-shattering kaboom!
     */
    log_err("failed to exec script '%s'", cmd);
    switch (errno) {
    case EACCES:
        html_err(cmd, "Permission denied");
        break;
    default:
        html_err(cmd, "Failed to start script");
        break;
    }
    exit(255);
}
